package de.V10lator.BukkitHTTPD;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.UUID;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.logging.Logger;

import org.bukkit.Server;
import org.bukkit.configuration.InvalidConfigurationException;
import org.bukkit.configuration.file.YamlConfiguration;
import org.bukkit.event.Event;
import org.bukkit.plugin.PluginDescriptionFile;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.scheduler.BukkitScheduler;

/**
 * 
 * @author codename_B
 * @author V10lator
 *
 */
public final class BukkitHTTPD extends JavaPlugin {

	/**
	 * The main configuration file.
	 */
	public static final String OLD_CONFIG_FILE = "server.properties";
	private final File f = new File("plugins/BukkitHTTPD.yml");
	final YamlConfiguration config = new YamlConfiguration();
	
	/**
	 * The save file.
	 */
	private final File sf = new File("plugins/BukkitHTTPD.sav");
	final YamlConfiguration save = new YamlConfiguration();
	/**
	 * The logging utility (used for error logging).
	 */
	private Logger log;
	
	/**
	 * The NanoHTTPD session
	 */
	private NanoHTTPD httpd;
	
	/**
	 * RAM and buffers between BukkitHTTPD and NanoHTTPD/V10parser.
	 */
	final RAM ram = new RAM();
	final HashMap<UUID, String> chatBuffer = new HashMap<UUID, String>();
	final HashMap<UUID, Long> chatMappings = new HashMap<UUID, Long>();
	
    /**
	 * The port and the directory of the HTTP daemon server.
	 */
	int httpdPort;
	static File wwwDir;

	public void onDisable()
	{
	  if(httpd != null)
	    httpd.stop();
	  getServer().getScheduler().cancelTasks(this);
	  saveConfig();
	  log.info("["+getDescription().getName()+"] disabled!");
	}

	public void onEnable()
	{
	  Server s = getServer();
	  log = s.getLogger();
	  String www;
	  try
	  {
		if(!f.exists())
		{
		  Properties props = new Properties();
		  props.load(new FileReader(OLD_CONFIG_FILE));
		  httpdPort = Integer.parseInt(props.getProperty("http-daemon-port", "8107"));
		  www = props.getProperty("http-dir", "plugins/BukkitHTTPD/");
		  ram.serverName = props.getProperty("server-name", "My Awesome Server");
		  ram.color = props.getProperty("text-color", "000000");
		  f.createNewFile();
		  config.load(f);
		  ram.hidePort = props.getProperty("hide-port", "false") != null;
		}
		else
		{
		  config.load(f);
		  httpdPort = config.getInt("Port", 8107);
		  www = config.getString("Directory", "plugins/BukkitHTTPD/");
		  ram.hidePort = config.getBoolean("HidePort", false);
		  ram.serverName = config.getString("ServerName", ram.serverName);
		  ram.color = config.getString("Color", ram.color);
		}
		if(!www.endsWith("/"))
		  www += "/";
		wwwDir = new File(www);
		try
		{
		  save.load(sf);
		}
		catch(FileNotFoundException e)
		{
		  sf.createNewFile();
		  save.load(sf);
		}
		for(String sp: save.getKeys(false))
		  ram.passwords.put(sp, save.getString(sp));
		saveConfig();
	  }
	  catch (IOException ex)
	  {
		log.info("["+getDescription().getName()+"] error!");
		ex.printStackTrace();
		s.getPluginManager().disablePlugin(this);
		return;
	  }
	  catch(NumberFormatException e)
	  {
		  log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		s.getPluginManager().disablePlugin(this);
		return;
	  }
	  catch(InvalidConfigurationException e)
	  {
		  log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		s.getPluginManager().disablePlugin(this);
		return;
	  }
	  wwwDir.mkdirs();
	  checkFiles();
	  BukkitScheduler bs = s.getScheduler();
	  bs.scheduleSyncRepeatingTask(this, new RAMFiller(this), 0L, 3L);
	  try
	  {
		httpd = new NanoHTTPD(this, httpdPort, wwwDir);
	  }
	  catch (IOException e)
	  {
		  log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		s.getPluginManager().disablePlugin(this);
		return;
	  }
	  bs.scheduleAsyncDelayedTask(this, new BukkitHTTPDThread(ram));
	  s.getPluginManager().registerEvent(Event.Type.PLAYER_CHAT, new BukkitHTTPDPlayerListener(this), Event.Priority.Monitor, this);
	  getCommand("httpd").setExecutor(new BukkitHTTPDCommandParser(this));
	  PluginDescriptionFile pdf = getDescription();
	  log.info("["+pdf.getName()+"] v"+pdf.getVersion()+" enabled!");
	}
	
	public void saveConfig()
	{
	  while(!ram.lock.compareAndSet(false, true))
	    continue;
	  config.set("HidePort", ram.hidePort);
	  config.set("Port", httpdPort);
	  config.set("Directory", wwwDir.toString());
	  config.set("ServerName", ram.serverName);
	  config.set("Color", ram.color);
	  config.set("MinecraftBans", ram.minecraftBans);
	  
	  //TODO: Enable after bugfix.
	  //for(String sk: save.getKeys(false))
		//save.set(sk, null);
	  
	  for(Entry<String, String> e: ram.passwords.entrySet())
		save.set(e.getKey(), e.getValue());
	  ram.lock.set(false);
	  try
	  {
	    config.save(f);
	    save.save(sf);
	  }
	  catch(IOException e)
	  {
		log.info("["+getDescription().getName()+"] could not save!");
	    e.printStackTrace();
	  }
	}
	
	public void reloadConfig()
	{
	  while(!ram.lock.compareAndSet(false, true))
		continue;
	  String www;
	  Server s = getServer();
	  try
	  {
		config.load(f);
		httpdPort = config.getInt("Port", 8107);
		www = config.getString("Directory", "plugins/BukkitHTTPD/");
		ram.hidePort = config.getBoolean("HidePort", false);
		ram.serverName = config.getString("ServerName", ram.serverName);
		ram.color = config.getString("Color", ram.color);
		if(!www.endsWith("/"))
		  www += "/";
		wwwDir = new File(www);
	  }
	  catch(FileNotFoundException e)
	  {
		Properties props = new Properties();
		try
		{
		  props.load(new FileReader(OLD_CONFIG_FILE));
		  httpdPort = Integer.parseInt(props.getProperty("http-daemon-port", "8107"));
		  www = props.getProperty("http-dir", "plugins/BukkitHTTPD/");
		  ram.serverName = props.getProperty("server-name", "My Awesome Server");
		  ram.color = props.getProperty("text-color", "000000");
		  ram.hidePort = props.getProperty("hide-port", "false") != null;
		  f.createNewFile();
		  config.load(f);
		}
		catch(FileNotFoundException e1)
		{
		  log.info("["+getDescription().getName()+"] error!");
		  e1.printStackTrace();
		  ram.lock.set(false);
		  s.getPluginManager().disablePlugin(this);
		  return;
		}
		catch(IOException e1)
		{
		  log.info("["+getDescription().getName()+"] error!");
		  e1.printStackTrace();
		  ram.lock.set(false);
		  s.getPluginManager().disablePlugin(this);
		  return;
		}
		catch(InvalidConfigurationException e1)
		{
		  log.info("["+getDescription().getName()+"] error!");
		  e1.printStackTrace();
		  ram.lock.set(false);
		  s.getPluginManager().disablePlugin(this);
		  return;
		}
		
	  }
	  catch(IOException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		ram.lock.set(false);
		s.getPluginManager().disablePlugin(this);
		return;
	  }
	  catch (InvalidConfigurationException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		ram.lock.set(false);
		s.getPluginManager().disablePlugin(this);
		return;
	  }
	  try
	  {
		save.load(sf);
		for(String sp: save.getKeys(false))
		  ram.passwords.put(sp, save.getString(sp));
	  }
	  catch(IOException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		ram.lock.set(false);
		s.getPluginManager().disablePlugin(this);
		return;
	  }
	  catch(InvalidConfigurationException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		ram.lock.set(false);
		s.getPluginManager().disablePlugin(this);
		return;
	  }
	  
	  ram.lock.set(false);
	  httpd.stop();
	  BukkitScheduler bs = s.getScheduler();
	  bs.cancelTasks(this);
	  saveConfig();
	  wwwDir.mkdirs();
	  checkFiles();
	  bs.scheduleSyncRepeatingTask(this, new RAMFiller(this), 0L, 3L);
	  try
	  {
		httpd = new NanoHTTPD(this, httpdPort, wwwDir);
	  }
	  catch (IOException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		s.getPluginManager().disablePlugin(this);
		return;
	  }
	  bs.scheduleAsyncDelayedTask(this, new BukkitHTTPDThread(ram));
	}
	
	public void checkFiles()
	{
	  String prefix = "["+getDescription().getName()+"] ";
	  if(!wwwDir.exists())
	  {
		if(wwwDir.mkdirs())
	      log.info(prefix+"www folder created.");
	    else
	    {
	      log.info(prefix+"can't create www folder!");
	      return;
	    }
	  }
	  
	  try
	  {
		JarFile jar = new JarFile(getFile());
		Enumeration<JarEntry> entries = jar.entries();
		String pkg = jar.getEntry(getClass().getPackage().getName().replace(".", "/")+"/www/").getName();
		File f;
		while(entries.hasMoreElements())
		{
		  JarEntry entry = entries.nextElement();
		  String en = entry.getName();
		  if(!en.startsWith(pkg))
			continue;
		  if(entry.isDirectory())
			new File(wwwDir, en.substring(pkg.length())).mkdirs();
		  else
		  {
			f = new File(wwwDir, en.substring(pkg.length()));
			if(!f.canRead())
			{
			  try
			  {
				if(!f.exists())
				{
				  f.createNewFile();
				  InputStream is = jar.getInputStream(entry);
				  FileOutputStream os = new FileOutputStream(f);
				  byte[] buf = new byte[4068];
				  is.read(buf, 0, 4068);
				  os.write(buf);
				  is.close();
				  os.close();
				}
			  }
			  catch (IOException e)
		      {
				log.info(prefix+"Can't create default file!");
			    e.printStackTrace();
		      }
			}
		  }
		}
	  }
	  catch(IOException e)
	  {
		e.printStackTrace();
		log.info(prefix+"Can't create default file!");
		e.printStackTrace();
	  }
}
}